/* This file is part of Zanshin

   Copyright 2014 Kevin Ottens <ervin@kde.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
   USA.
*/


#include "akonadiartifactqueries.h"

#include "akonadicollectionfetchjobinterface.h"
#include "akonadiitemfetchjobinterface.h"
#include "akonadimonitorimpl.h"
#include "akonadiserializer.h"
#include "akonadistorage.h"

#include "utils/jobhandler.h"

using namespace Akonadi;

ArtifactQueries::ArtifactQueries(QObject *parent)
    : QObject(parent),
      m_storage(new Storage),
      m_serializer(new Serializer),
      m_monitor(new MonitorImpl),
      m_fetchContentTypeFilter(StorageInterface::Tasks|StorageInterface::Notes),
      m_ownInterfaces(true)
{
    connect(m_monitor, SIGNAL(itemAdded(Akonadi::Item)), this, SLOT(onItemAdded(Akonadi::Item)));
    connect(m_monitor, SIGNAL(itemRemoved(Akonadi::Item)), this, SLOT(onItemRemoved(Akonadi::Item)));
    connect(m_monitor, SIGNAL(itemChanged(Akonadi::Item)), this, SLOT(onItemChanged(Akonadi::Item)));
    connect(m_monitor, SIGNAL(collectionSelectionChanged(Akonadi::Collection)), this, SLOT(onCollectionSelectionChanged()));
}

ArtifactQueries::ArtifactQueries(StorageInterface *storage, SerializerInterface *serializer, MonitorInterface *monitor)
    : m_storage(storage),
      m_serializer(serializer),
      m_monitor(monitor),
      m_fetchContentTypeFilter(StorageInterface::Tasks|StorageInterface::Notes),
      m_ownInterfaces(false)
{
    connect(m_monitor, SIGNAL(itemAdded(Akonadi::Item)), this, SLOT(onItemAdded(Akonadi::Item)));
    connect(m_monitor, SIGNAL(itemRemoved(Akonadi::Item)), this, SLOT(onItemRemoved(Akonadi::Item)));
    connect(m_monitor, SIGNAL(itemChanged(Akonadi::Item)), this, SLOT(onItemChanged(Akonadi::Item)));
    connect(m_monitor, SIGNAL(collectionSelectionChanged(Akonadi::Collection)), this, SLOT(onCollectionSelectionChanged()));
}

ArtifactQueries::~ArtifactQueries()
{
    if (m_ownInterfaces) {
        delete m_storage;
        delete m_serializer;
        delete m_monitor;
    }
}

void ArtifactQueries::setApplicationMode(ArtifactQueries::ApplicationMode mode)
{
    if (mode == TasksOnly) {
        m_fetchContentTypeFilter = StorageInterface::Tasks;
    } else if (mode == NotesOnly) {
        m_fetchContentTypeFilter = StorageInterface::Notes;
    }
}

ArtifactQueries::ArtifactResult::Ptr ArtifactQueries::findInboxTopLevel() const
{
    if (!m_findInbox) {
        {
            ArtifactQueries *self = const_cast<ArtifactQueries*>(this);
            self->m_findInbox = self->createArtifactQuery();
        }

        m_findInbox->setFetchFunction([this] (const ArtifactQuery::AddFunction &add) {
            CollectionFetchJobInterface *job = m_storage->fetchCollections(Akonadi::Collection::root(),
                                                                           StorageInterface::Recursive,
                                                                           m_fetchContentTypeFilter);
            Utils::JobHandler::install(job->kjob(), [this, job, add] {
                if (job->kjob()->error() != KJob::NoError)
                    return;

                for (auto collection : job->collections()) {
                    if (!m_serializer->isSelectedCollection(collection))
                        continue;

                    ItemFetchJobInterface *job = m_storage->fetchItems(collection);
                    Utils::JobHandler::install(job->kjob(), [this, job, add, collection] {
                        if (job->kjob()->error() != KJob::NoError)
                            return;

                        for (auto item : job->items()) {
                            //We have to set the parent to since we rely on attributes being available in isSelectedCollection
                            item.setParentCollection(collection);
                            add(item);
                        }
                    });
                }
            });
        });

        m_findInbox->setConvertFunction([this] (const Akonadi::Item &item) -> Domain::Artifact::Ptr {
            if (m_serializer->isTaskItem(item)) {
                auto task = m_serializer->createTaskFromItem(item);
                return Domain::Artifact::Ptr(task);

            } else if (m_serializer->isNoteItem(item)) {
                auto note = m_serializer->createNoteFromItem(item);
                return Domain::Artifact::Ptr(note);

            } else {
                return Domain::Artifact::Ptr();
            }
        });

        m_findInbox->setUpdateFunction([this] (const Akonadi::Item &item, Domain::Artifact::Ptr &artifact) {
            if (auto task = artifact.dynamicCast<Domain::Task>()) {
                m_serializer->updateTaskFromItem(task, item);
            } else if (auto note = artifact.dynamicCast<Domain::Note>()) {
                m_serializer->updateNoteFromItem(note, item);
            }
        });

        m_findInbox->setPredicateFunction([this] (const Akonadi::Item &item) -> bool {
            const bool isTask = m_serializer->isTaskItem(item);
            const bool isNote = m_serializer->isNoteItem(item);
            const bool excluded = !m_serializer->relatedUidFromItem(item).isEmpty()
                               || (!m_serializer->isTaskItem(item) && !m_serializer->isNoteItem(item))
                               || (m_serializer->isTaskItem(item) && m_serializer->hasContextTags(item))
                               || m_serializer->hasAkonadiTags(item)
                               || !m_serializer->isSelectedCollection(item.parentCollection())
                               || (m_fetchContentTypeFilter & StorageInterface::Tasks) && !isTask
                               || (m_fetchContentTypeFilter & StorageInterface::Notes) && !isNote;

            return !excluded;
        });

        m_findInbox->setRepresentsFunction([this] (const Akonadi::Item &item, const Domain::Artifact::Ptr &artifact) -> bool {
            return m_serializer->representsItem(artifact, item);
        });
    }

    return m_findInbox->result();
}

ArtifactQueries::TagResult::Ptr ArtifactQueries::findTags(Domain::Artifact::Ptr artifact) const
{
    Q_UNUSED(artifact);
    qFatal("Not implemented yet");
    return TagResult::Ptr();
}

void ArtifactQueries::onItemAdded(const Item &item)
{
    foreach (const ArtifactQuery::Ptr &query, m_artifactQueries)
        query->onAdded(item);
}

void ArtifactQueries::onItemRemoved(const Item &item)
{
    foreach (const ArtifactQuery::Ptr &query, m_artifactQueries)
        query->onRemoved(item);
}

void ArtifactQueries::onItemChanged(const Item &item)
{
    foreach (const ArtifactQuery::Ptr &query, m_artifactQueries)
        query->onChanged(item);
}

void ArtifactQueries::onCollectionSelectionChanged()
{
    foreach (const ArtifactQuery::Ptr &query, m_artifactQueries)
        query->reset();
}

ArtifactQueries::ArtifactQuery::Ptr ArtifactQueries::createArtifactQuery()
{
    auto query = ArtifactQuery::Ptr::create();
    m_artifactQueries << query;
    return query;
}
