/* This file is part of Zanshin

   Copyright 2014 Kevin Ottens <ervin@kde.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
   USA.
*/


#include "applicationmodel.h"

#include "domain/artifactqueries.h"
#include "domain/contextqueries.h"
#include "domain/contextrepository.h"
#include "domain/projectqueries.h"
#include "domain/projectrepository.h"
#include "domain/noterepository.h"
#include "domain/tagqueries.h"
#include "domain/tagrepository.h"
#include "domain/taskqueries.h"
#include "domain/taskrepository.h"
#include "domain/datasourcerepository.h"

#include "presentation/artifacteditormodel.h"
#include "presentation/availablepagesmodel.h"
#include "presentation/availablesourcesmodel.h"
#include "presentation/datasourcelistmodel.h"

#include "utils/dependencymanager.h"

using namespace Presentation;

ApplicationModel::ApplicationModel(QObject *parent, ApplicationMode mode)
    : QObject(parent),
      m_availableSources(0),
      m_availablePages(0),
      m_currentPage(0),
      m_editor(0),
      m_artifactQueries(Utils::DependencyManager::globalInstance().create<Domain::ArtifactQueries>()),
      m_projectQueries(Utils::DependencyManager::globalInstance().create<Domain::ProjectQueries>()),
      m_projectRepository(Utils::DependencyManager::globalInstance().create<Domain::ProjectRepository>()),
      m_contextQueries(Utils::DependencyManager::globalInstance().create<Domain::ContextQueries>()),
      m_contextRepository(Utils::DependencyManager::globalInstance().create<Domain::ContextRepository>()),
      m_sourceQueries(Utils::DependencyManager::globalInstance().create<Domain::DataSourceQueries>()),
      m_sourceRepository(Utils::DependencyManager::globalInstance().create<Domain::DataSourceRepository>()),
      m_taskQueries(Utils::DependencyManager::globalInstance().create<Domain::TaskQueries>()),
      m_taskRepository(Utils::DependencyManager::globalInstance().create<Domain::TaskRepository>()),
      m_taskSourcesModel(0),
      m_noteRepository(Utils::DependencyManager::globalInstance().create<Domain::NoteRepository>()),
      m_noteSourcesModel(0),
      m_tagQueries(Utils::DependencyManager::globalInstance().create<Domain::TagQueries>()),
      m_tagRepository(Utils::DependencyManager::globalInstance().create<Domain::TagRepository>()),
      m_relationQueries(Utils::DependencyManager::globalInstance().create<Domain::RelationQueries>()),
      m_relationRepository(Utils::DependencyManager::globalInstance().create<Domain::RelationRepository>()),
      m_ownInterface(true),
      m_mode(mode)
{
    MetaTypes::registerAll();
    m_sourceQueries->setApplicationMode(m_mode == NotesOnly ? Domain::DataSourceQueries::NotesOnly : Domain::DataSourceQueries::TasksOnly);
    m_artifactQueries->setApplicationMode(m_mode == NotesOnly ? Domain::ArtifactQueries::NotesOnly : Domain::ArtifactQueries::TasksOnly);
    m_tagQueries->setApplicationMode(m_mode == NotesOnly ? Domain::TagQueries::NotesOnly : Domain::TagQueries::TasksOnly);
    m_sourceRepository->setApplicationMode(m_mode == NotesOnly ? Domain::DataSourceRepository::NotesOnly : Domain::DataSourceRepository::TasksOnly);
}

ApplicationModel::ApplicationModel(Domain::ArtifactQueries *artifactQueries,
                                   Domain::ProjectQueries *projectQueries,
                                   Domain::ProjectRepository *projectRepository,
                                   Domain::ContextQueries *contextQueries,
                                   Domain::ContextRepository *contextRepository,
                                   Domain::DataSourceQueries *sourceQueries,
                                   Domain::DataSourceRepository *sourceRepository,
                                   Domain::TaskQueries *taskQueries,
                                   Domain::TaskRepository *taskRepository,
                                   Domain::NoteRepository *noteRepository,
                                   Domain::TagQueries *tagQueries,
                                   Domain::TagRepository *tagRepository,
                                   QObject *parent, ApplicationMode mode)
    : QObject(parent),
      m_availableSources(0),
      m_availablePages(0),
      m_currentPage(0),
      m_editor(0),
      m_artifactQueries(artifactQueries),
      m_projectQueries(projectQueries),
      m_projectRepository(projectRepository),
      m_contextQueries(contextQueries),
      m_contextRepository(contextRepository),
      m_sourceQueries(sourceQueries),
      m_sourceRepository(sourceRepository),
      m_taskQueries(taskQueries),
      m_taskRepository(taskRepository),
      m_taskSourcesModel(0),
      m_noteRepository(noteRepository),
      m_noteSourcesModel(0),
      m_tagQueries(tagQueries),
      m_tagRepository(tagRepository),
      m_ownInterface(false),
      m_mode(mode)
{
    MetaTypes::registerAll();
    m_sourceQueries->setApplicationMode(m_mode == NotesOnly ? Domain::DataSourceQueries::NotesOnly : Domain::DataSourceQueries::TasksOnly);
    m_artifactQueries->setApplicationMode(m_mode == NotesOnly ? Domain::ArtifactQueries::NotesOnly : Domain::ArtifactQueries::TasksOnly);
    m_tagQueries->setApplicationMode(m_mode == NotesOnly ? Domain::TagQueries::NotesOnly : Domain::TagQueries::TasksOnly);
}

ApplicationModel::~ApplicationModel()
{
    if (m_ownInterface) {
        delete m_artifactQueries;
        delete m_projectQueries;
        delete m_projectRepository;
        delete m_contextQueries;
        delete m_contextRepository;
        delete m_sourceQueries;
        delete m_sourceRepository;
        delete m_taskQueries;
        delete m_taskRepository;
        delete m_noteRepository;
        delete m_tagQueries;
        delete m_tagRepository;
    }
}

QAbstractItemModel *ApplicationModel::noteSourcesModel()
{
    if (!m_noteSourcesModel) {
        m_noteSourcesModel = new DataSourceListModel([this] { return noteSources(); }, this);
    }

    return m_noteSourcesModel;
}

Domain::QueryResult<Domain::DataSource::Ptr>::Ptr ApplicationModel::noteSources()
{
    if (!m_noteSources) {
        m_noteSources = m_sourceQueries->findNotes();
        m_noteSources->addPostInsertHandler([this](const Domain::DataSource::Ptr &source, int) {
            if (m_noteRepository->isDefaultSource(source)) {
                emit defaultNoteDataSourceChanged(source);
            }
        });
    }

    return m_noteSources;
}

Domain::DataSource::Ptr ApplicationModel::defaultNoteDataSource()
{
    QList<Domain::DataSource::Ptr> sources = noteSources()->data();

    if (sources.isEmpty())
        return Domain::DataSource::Ptr();

    auto source = std::find_if(sources.begin(), sources.end(),
                               [this] (const Domain::DataSource::Ptr &source) {
                                   return m_noteRepository->isDefaultSource(source);
                               });

    if (source != sources.end())
        return *source;
    else
        return sources.first();
}

QAbstractItemModel *ApplicationModel::taskSourcesModel()
{
    if (!m_taskSourcesModel) {
        m_taskSourcesModel = new DataSourceListModel([this] { return taskSources(); }, this);
    }

    return m_taskSourcesModel;
}

Domain::QueryResult<Domain::DataSource::Ptr>::Ptr ApplicationModel::taskSources()
{
    if (!m_taskSources) {
        m_taskSources = m_sourceQueries->findTasks();
    }

    return m_taskSources;
}

Domain::DataSource::Ptr ApplicationModel::defaultTaskDataSource()
{
    QList<Domain::DataSource::Ptr> sources = taskSources()->data();

    if (sources.isEmpty())
        return Domain::DataSource::Ptr();

    auto source = std::find_if(sources.begin(), sources.end(),
                               [this] (const Domain::DataSource::Ptr &source) {
                                   return m_taskRepository->isDefaultSource(source);
                               });

    if (source != sources.end())
        return *source;
    else
        return sources.first();
}

QObject *ApplicationModel::availableSources()
{
    if (!m_availableSources) {
        m_availableSources = new AvailableSourcesModel(m_sourceQueries,
                                                       m_sourceRepository,
                                                       this);
    }
    return m_availableSources;
}

QObject *ApplicationModel::availablePages()
{
    if (!m_availablePages) {
        m_availablePages = new AvailablePagesModel(m_artifactQueries,
                                                   m_projectQueries,
                                                   m_projectRepository,
                                                   m_contextQueries,
                                                   m_contextRepository,
                                                   m_taskQueries,
                                                   m_taskRepository,
                                                   m_noteRepository,
                                                   m_tagQueries,
                                                   m_tagRepository,
                                                   this,
                                                   m_mode == NotesOnly ? AvailablePagesModel::NotesOnly : AvailablePagesModel::TasksOnly);
    }
    return m_availablePages;
}

QObject *ApplicationModel::currentPage()
{
    return m_currentPage;
}

QObject *ApplicationModel::editor()
{
    if (!m_editor) {
        m_editor = new ArtifactEditorModel(m_taskRepository, m_noteRepository, m_relationQueries, m_relationRepository, m_tagQueries, m_tagRepository, this);
    }

    return m_editor;
}

void ApplicationModel::setCurrentPage(QObject *page)
{
    if (page == m_currentPage)
        return;

    m_currentPage = page;
    emit currentPageChanged(page);
}

void ApplicationModel::setDefaultNoteDataSource(Domain::DataSource::Ptr source)
{
    m_noteRepository->setDefaultSource(source);
}

void ApplicationModel::setDefaultTaskDataSource(Domain::DataSource::Ptr source)
{
    m_taskRepository->setDefaultSource(source);
}
