/* This file is part of Zanshin

   Copyright 2014 Kevin Ottens <ervin@kde.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
   USA.
*/


#ifndef PRESENTATION_ARTIFACTEDITORMODEL_H
#define PRESENTATION_ARTIFACTEDITORMODEL_H

#include <QDateTime>
#include <QObject>

#include "domain/task.h"
#include "domain/relation.h"
#include "domain/tag.h"

class QTimer;

namespace Domain {
    class NoteRepository;
    class TaskRepository;
    class RelationQueries;
    class RelationRepository;
    class TagQueries;
    class TagRepository;
}

namespace Presentation {

class ArtifactEditorModel : public QObject
{
    Q_OBJECT
    Q_PROPERTY(Domain::Artifact::Ptr artifact READ artifact WRITE setArtifact NOTIFY artifactChanged)
    Q_PROPERTY(QString text READ text WRITE setText NOTIFY textChanged)
    Q_PROPERTY(QString title READ title WRITE setTitle NOTIFY titleChanged)
    Q_PROPERTY(QDateTime startDate READ startDate WRITE setStartDate NOTIFY startDateChanged)
    Q_PROPERTY(QDateTime dueDate READ dueDate WRITE setDueDate NOTIFY dueDateChanged)
    Q_PROPERTY(QString delegateText READ delegateText NOTIFY delegateTextChanged)
    Q_PROPERTY(bool hasTaskProperties READ hasTaskProperties NOTIFY hasTaskPropertiesChanged)
    Q_PROPERTY(int progress READ progress WRITE setProgress NOTIFY progressChanged)
    Q_PROPERTY(int status READ status WRITE setStatus NOTIFY statusChanged)
    Q_PROPERTY(QList<Domain::Relation::Ptr> relations READ relations NOTIFY relationsChanged)
    Q_PROPERTY(Domain::Tag::List tags READ tags NOTIFY tagsChanged)
    Q_PROPERTY(Domain::Artifact::Attachment::List attachments READ attachments NOTIFY attachmentsChanged)
    Q_PROPERTY(Domain::Recurrence::Ptr recurrence READ recurrence WRITE setRecurrence NOTIFY recurrenceChanged)
    Q_PROPERTY(Domain::Alarm::Ptr alarm READ alarm READ alarm WRITE setAlarm NOTIFY alarmChanged)
    Q_PROPERTY(bool saveNeeded READ isSaveNeeded NOTIFY saveNeededChanged)
public:
    explicit ArtifactEditorModel(Domain::TaskRepository *taskRepository,
                                 Domain::NoteRepository *noteRepository,
                                 Domain::RelationQueries *relationQueries = 0,
                                 Domain::RelationRepository *relationRepository = 0,
                                 Domain::TagQueries *tagQueries = 0,
                                 Domain::TagRepository *tagRepository = 0,
                                 QObject *parent = 0);
    ~ArtifactEditorModel();

    Domain::Artifact::Ptr artifact() const;
    void setArtifact(const Domain::Artifact::Ptr &artifact);

    bool hasTaskProperties() const;

    QString text() const;
    QString title() const;
    QDateTime startDate() const;
    QDateTime dueDate() const;
    QString delegateText() const;
    int progress() const;
    int status() const;
    QList<Domain::Relation::Ptr> relations() const;
    Domain::Tag::List tags() const;
    Domain::Artifact::Attachment::List attachments() const;
    Domain::Recurrence::Ptr recurrence() const;
    Domain::Alarm::Ptr alarm() const;

    static int autoSaveDelay();

public slots:
    void setText(const QString &text);
    void setTitle(const QString &title);
    void setStartDate(const QDateTime &start);
    void setDueDate(const QDateTime &due);
    void setDelegate(const QString &name, const QString &email);
    void delegate(const QString &name, const QString &email);
    void setProgress(int progress);
    void setStatus(int status);
    void removeRelation(const Domain::Relation::Ptr &);
    void setRecurrence(const Domain::Recurrence::Ptr&);
    void removeAttachment(const Domain::Artifact::Attachment::Ptr &);
    void addAttachment(const Domain::Artifact::Attachment::Ptr &);
    void removeTag(const Domain::Tag::Ptr &);
    void addTag(const QString &);
    void addTag(const Domain::Tag::Ptr &);

    void setFrequency(Domain::Recurrence::Frequency frequency, int intervall);
    void setRepeatEnd(QDateTime endDate);
    void setRepeatEnd(int count);
    void setRepeatEndless();
    void setNoRepeat();
    void setExceptionDates(const QList<QDateTime> &exceptionDates);
    void setByDay(const QList< Domain::Recurrence::Weekday > &dayList);
    void setByDayPosition(Domain::Recurrence::WeekPosition position);
    void setByMonth(const QList< int > &monthList);
    void setByMonthDays(const QList< int > &dayList);

    void setAlarm(const Domain::Alarm::Ptr &alarm);
    void setAlarmType(int type);
    void setAlarmPosition(int position);
    void setAlarmOffset(int offset);

    void save();

signals:
    void artifactChanged(const Domain::Artifact::Ptr &artifact);
    void hasTaskPropertiesChanged(bool hasTaskProperties);
    void textChanged(const QString &text);
    void titleChanged(const QString &title);
    void startDateChanged(const QDateTime &date);
    void dueDateChanged(const QDateTime &due);
    void delegateTextChanged(const QString &delegateText);
    void progressChanged(int progress);
    void statusChanged(int status);
    void relationsChanged(const QList<Domain::Relation::Ptr> &relations);
    void attachmentsChanged(const Domain::Artifact::Attachment::List &attachments);
    void recurrenceChanged(const Domain::Recurrence::Ptr &recurrence);
    void alarmChanged(const Domain::Alarm::Ptr &alarm);
    void tagsChanged(const Domain::Tag::List &tags);

    void saveNeededChanged(bool saveNeeded);

private slots:
    void onTextChanged(const QString &text);
    void onTitleChanged(const QString &title);
    void onStartDateChanged(const QDateTime &start);
    void onDueDateChanged(const QDateTime &due);
    void onDelegateChanged(const Domain::Task::Delegate &delegate);
    void onProgressChanged(int status);
    void onStatusChanged(int status);
    void onRecurrenceChanged(const Domain::Recurrence::Ptr &recurrence);
    void onAttachmentsChanged(const Domain::Artifact::Attachment::List &attachments);
    void onAlarmChanged(const Domain::Alarm::Ptr &alarm);
    void onAlarmsChanged(const Domain::Alarm::List &alarms);


private:
    void setSaveNeeded(bool needed);
    bool isSaveNeeded() const;


    Domain::TaskRepository *m_taskRepository;
    Domain::NoteRepository *m_noteRepository;
    Domain::RelationQueries *m_relationQueries;
    Domain::RelationRepository *m_relationRepository;
    Domain::TagQueries *m_tagQueries;
    Domain::TagRepository *m_tagRepository;

    Domain::Artifact::Ptr m_artifact;

    QString m_text;
    QString m_title;
    QDateTime m_start;
    QDateTime m_due;
    Domain::Task::Delegate m_delegate;
    int m_progress;
    Domain::Task::Status m_status;
    QList<Domain::Relation::Ptr> m_relations;
    QList<Domain::Tag::Ptr> m_tags;
    Domain::Recurrence::Ptr m_recurrence;
    Domain::Artifact::Attachment::List m_attachments;
    Domain::Alarm::Ptr m_alarm;

    bool m_saveNeeded;
};

}

#endif // PRESENTATION_ARTIFACTEDITORMODEL_H
